import { ActivityOptions, Client, Message } from "discord.js";
import { Console } from "../console/console";
import { ActivityLoader } from "../data/activityLoader";
import { Response } from "../response/response";
import { VoiceOptions, DJCommands, SongOptions } from "./clientVoice";
export interface MessageResponse {
    message: string;
    react?: string;
    reply?: boolean;
    audio?: boolean;
    refIndex: string;
}
export interface SBotOptions {
    token: string | undefined;
    activityRefreshInterval?: number;
    logLocation?: string;
    ignorePrivacy?: boolean;
    logIncomingMessage?: boolean;
}
export declare class SBotClient {
    client: Client;
    get user(): import("discord.js").ClientUser | null;
    private utility;
    private options;
    /**
     * @param options Options has 5 fields
     *
     * - token: Bot Token, Defaults to `process.env.DISCORD_TOKEN`
     *
     * - activityRefreshInterval (Optional): Default 600 seconds
     *
     * - logLocation: Location for logging, Default is deactivated
     *
     * - ignorePrivacy: If set to true, bot will not deafen when joining voice channel,
     * PS: this bot does not listen to voice chat
     *
     * - logIncomingMessage: If set to true, will report every incoming message
     * just like older version of Salim Bot
     */
    constructor(options?: SBotOptions);
    private props;
    useComputedActivity(activity: ActivityOptions): void;
    useActivities(loader: ActivityLoader): void;
    private setSelfActivity;
    useResponse(response: Response): void;
    private response;
    useConsole(console: Console): void;
    private voiceOptions?;
    private voiceCtrl?;
    /**
     * Activate voice feature, you will need to provide words
     * for your bot to speak under each circumstance.
     *
     * See: src/client/clientVoice.ts @ `VoiceOptions`
     */
    useVoice(options: VoiceOptions): void;
    private djCorgiResponse;
    private corgiSwiftQueue;
    private corgiSwiftJutsu;
    private corgiSwiftClearQueue;
    private printSongsList;
    private songOptions?;
    private djCommands?;
    /**
     * Activate DJ Feature, see [Salim Bot](https://github.com/Leomotors/Salim-Bot)
     *
     * Basically, like other part of this framework, you need to provide all
     * information and the framework will take care of them
     */
    useDJ(SongOptions: SongOptions[], Options: DJCommands): void;
    requestSong(msg: Message): Promise<void>;
    private corgiPlaySong;
}
